#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from threading import Thread, current_thread
import time
from q2gui.q2form import Q2Form
import q2gui.q2app as q2app


def center_window(form: Q2Form):
    w, h = q2app.q2_app.get_size()
    h -= q2app.q2_app.get_stdout_height()

    form.form_stack[0].set_size(int(w * 0.5), int(h * 0.5))
    form.form_stack[0].set_position(int(w * 0.25), int(h * 0.15))


def q2Mess(mess="", title="Message"):
    form = Q2Form(title)
    form.do_not_save_geometry = True
    form.add_control("/v")
    form.add_control("mess", control="text", data=f"{mess}", readonly=True)
    if form.add_control("/h"):
        form.add_control("/s")

        form.add_control(
            "_ok_button",
            "Ok",
            control="button",
            eat_enter=True,
            hotkey="PgDown",
            valid=form.close,
            tag="_ok_button"
        )

        form.add_control("/s")
        form.add_control("/")

    def after_form_show(form=form):
        center_window(form)
        form.w._ok_button.set_focus()

    form.after_form_show = after_form_show

    form.show_app_modal_form()
    q2app.q2_app.process_events()


q2_mess = q2mess = q2Mess


def q2AskYN(mess, title="Question"):
    form = Q2Form(title)
    form.do_not_save_geometry = True
    form.choice = 0
    form.add_control("/")

    form.controls.add_control("mess", control="text", data=f"{mess}", readonly="*", disabled="")

    if form.add_control("/h"):
        form.add_control("/s")

        def buttonPressed(form=form, answer=0):
            form.choice = answer
            form.close()

        form.add_control(
            "cancel",
            "Cancel",
            control="button",
            valid=lambda: buttonPressed(form, 1),
            eat_enter="*",
        )
        form.add_control(
            "ok",
            "Ok",
            control="button",
            valid=lambda: buttonPressed(form, 2),
            eat_enter="*",
            tag="ok",
        )
        form.add_control("/s")
        form.add_control("/")

    def after_form_show(form=form):
        center_window(form)
        form.w.cancel.set_focus()

    form.after_form_show = after_form_show
    form.show_app_modal_form()
    q2app.q2_app.process_events()
    return form.choice


q2_ask = q2ask = q2_ask_yn = q2askyn = q2AskYN


class Q2Thread(Thread):
    def __init__(self, group=None, target=None, name=None, args=(), kwargs={}):
        Thread.__init__(self, group, target, name, args, kwargs)
        self._target = target
        self._args = args
        self.min = 0
        self.max = 0
        self.shadow_value = 0
        self.value = 0
        self._return = None
        self._exc = None
        self.start_time = time.time()

    @staticmethod
    def set_min(value):
        current_thread().min = value

    @staticmethod
    def set_max(value):
        current_thread().max = value

    @staticmethod
    def get_max():
        c_thread = current_thread()
        if hasattr(c_thread, "max"):
            return c_thread.max
        else:
            return 0

    @staticmethod
    def step(step_value=1):
        current_thread().shadow_value += 1
        sv = current_thread().shadow_value
        if sv % step_value == 0:
            current_thread().value = sv

    @staticmethod
    def get_current():
        return current_thread().value

    def time(self):
        return time.time() - self.start_time

    def run(self):
        try:
            self._return = self._target(*self._args)
        except Exception as e:
            self._exc = e


class Q2WaitForm:
    def __init__(self, mess, worker_thread):
        self.tick = {}
        self.worker_thread = worker_thread
        self.wait_window = Q2Form("Wait...")
        self.wait_window.do_not_save_geometry = True
        self.wait_window.add_control("", label=mess, control="label")
        # self.wait_window.add_control("label", label="", data=mess)
        steps_count_separator = ""
        if Q2Thread.get_max() != 0:
            steps_count_separator = "/"

        if self.wait_window.add_control("/h"):
            self.wait_window.add_control("progressbar", "", control="progressbar")
            self.wait_window.add_control("min", "", control="label")
            self.wait_window.add_control("", steps_count_separator, control="label")
            self.wait_window.add_control("value", "", control="label")
            self.wait_window.add_control("", steps_count_separator, control="label")
            self.wait_window.add_control("max", "", control="label")
            self.wait_window.add_control("time", "", control="label")
        self.wait_window.add_control("/")
        self.show()

        if self.worker_thread.min != 0:
            self.wait_window.w.progressbar.set_min(self.worker_thread.min)
            self.wait_window.s.min = self.worker_thread.min
        if self.worker_thread.max != 0:
            self.wait_window.w.progressbar.set_max(self.worker_thread.max)
            self.wait_window.s.max = self.worker_thread.max

    def step(self):
        if self.worker_thread.value != 0:
            self.wait_window.w.progressbar.set_value(self.worker_thread.value)
            self.wait_window.s.value = self.worker_thread.value

        thread_time = int(self.worker_thread.time())
        # print(self.worker_thread.time())
        sec = thread_time % 60
        min = (thread_time - sec) % 3600
        hours = thread_time - min * 3600 - sec
        sec = int(sec)
        self.wait_window.s.time = f" Time {hours:02}:{min:02}:{sec:02}"
        q2app.q2_app.process_events()

    def show(self):
        self.wait_window.show_mdi_form()
        q2app.q2_app.process_events()
        w = q2app.q2_app.get_size()[0]
        fh = self.wait_window.form_stack[0].get_size()[1]
        self.wait_window.form_stack[0].set_size(int(w * 0.9), fh)
        left, top = self.wait_window.form_stack[0].center_pos()
        self.wait_window.form_stack[0].set_position(left, top)
        q2app.q2_app.process_events()

    def close(self):
        self.wait_window.close()
        q2app.q2_app.process_events()


def q2WaitStep(step_value=1):
    Q2Thread.step(step_value)


def q2WaitMax(max_value=0):
    Q2Thread.set_max(max_value)


def q2working(worker, mess=""):
    wait_window = None
    wait_window_on = False
    last_focus_widget = q2app.q2_app.focus_widget()
    last_progressbar_value = 0
    last_progressbar_time = 0
    q2app.q2_app.lock()
    worker_thread = Q2Thread(target=worker)
    worker_thread.start()
    while worker_thread.is_alive():
        time.sleep(0.3)
        if worker_thread.time() > 1 and wait_window_on is not True:
            wait_window_on = True
            wait_window = Q2WaitForm(mess, worker_thread)
        if wait_window_on is True:
            if worker_thread.min < worker_thread.max:
                if worker_thread.value != 0 and last_progressbar_value != worker_thread.value:
                    wait_window.step()
            elif worker_thread.time() - last_progressbar_time > 1:
                wait_window.step()
                last_progressbar_time = worker_thread.time()
        last_progressbar_value = worker_thread.value
        q2app.q2_app.process_events()
    q2app.q2_app.unlock()
    if wait_window is not None:
        wait_window.close()
    if hasattr(last_focus_widget, "set_focus"):
        last_focus_widget.set_focus()

    q2app.q2_app.show_statusbar_mess(f"{worker_thread.time():.2f}")
    if worker_thread._exc:
        raise worker_thread._exc
    return worker_thread._return


q2Wait = q2working


class Q2WaitShow:
    def __init__(self, *args):
        self.interrupted = False
        mess = "Working... \t"
        max_range = 0
        for x in args:
            x = str(x)
            if x.isdigit():
                max_range = int(x)
            else:
                mess = x
        self.mess = mess
        self.interval = 0.5 if max_range > 100 or max_range == 0 else 0

        self.wait_window = Q2Form("Wait...")
        self.wait_window.do_not_save_geometry = True
        self.wait_window.add_control("mess", label=self.mess, control="label", alignment=5)
        steps_count_separator = ">"

        if self.wait_window.add_control("/h"):
            self.wait_window.add_control("progressbar", "", control="progressbar")
            self.wait_window.add_control("value", "", control="label")
            self.wait_window.add_control("", steps_count_separator, control="label")
            self.wait_window.add_control("max", "", control="label")
            self.wait_window.add_control("time", "", control="label")
        self.wait_window.add_control("/")

        q2app.q2_app.disable_toolbar(True)
        q2app.q2_app.disable_menubar(True)
        q2app.q2_app.disable_tabbar(True)

        self.show()
        self.start_time = time.time()
        self.last_time = time.time()

        self.value = -1
        self.step()
        self.wait_window.w.progressbar.set_min(self.value)
        self.wait_window.s.min = 0
        self.wait_window.w.progressbar.set_max(max_range)
        self.wait_window.s.max = max_range if max_range else ""
        q2app.q2_app.process_events()
        self.last_focus_widget = q2app.q2_app.focus_widget()
        self.wait_window.after_form_closed = self.wait_windows_after_form_closed

    def wait_windows_after_form_closed(self):
        self.interrupted = True

    def step(self, *args):
        """
        
        """
        if self.interrupted:
            return True
        text = ""
        for x in args:
            x = str(x)
            if x.isdigit():
                self.interval = int(x) / 1000
            else:
                text = x

        self.value += 1
        if self.wait_window.w.progressbar:
            if self.interval and (time.time() - self.last_time) < self.interval:
                return
            else:
                self.last_time = time.time()

            self.wait_window.w.progressbar.set_value(self.value)
            self.wait_window.s.value = self.value
            self.wait_window.s.mess = f"{self.mess}{text}"

            thread_time = int(time.time() - self.start_time)
            sec = thread_time % 60
            min = int((thread_time - sec) / 60)
            sec = int(sec)
            self.wait_window.s.time = f" {min}:{sec:02}"

            q2app.q2_app.process_events()
        else:
            q2app.q2_app.process_events()

    def show(self):
        q2app.q2_app.disable_current_form()
        self.wait_window.show_mdi_form()
        q2app.q2_app.process_events()
        w = q2app.q2_app.get_size()[0]
        fh = self.wait_window.form_stack[0].get_size()[1]
        self.wait_window.form_stack[0].set_size(int(w * 0.9), fh)
        left, top = self.wait_window.form_stack[0].center_pos()
        self.wait_window.form_stack[0].set_position(left, top)

        if hasattr(self.wait_window.prev_form, "name") and self.wait_window.prev_form.name == "Wait...":
            mo_h = self.wait_window.prev_form.form_stack[0].get_size()[1]
            self.wait_window.form_stack[0].hide_border()
            self.wait_window.form_stack[0].move_window(0, mo_h)

        q2app.q2_app.process_events()

    def close(self):
        if not self.interrupted:
            self.wait_window.close()
            q2app.q2_app.disable_current_form(False)
            if hasattr(self.last_focus_widget, "set_focus"):
                self.last_focus_widget.set_focus()
        q2app.q2_app.disable_toolbar(False)
        q2app.q2_app.disable_menubar(False)
        q2app.q2_app.disable_tabbar(False)
        q2app.q2_app.process_events()
        return self.value, time.time() - self.start_time


q2_wait_show = q2waitshow = q2wait = q2_wait = Q2WaitShow



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sys



from PyQt6.QtWidgets import QLabel
from PyQt6.QtCore import Qt

from q2gui.pyqt6.q2widget import Q2Widget

vertical_align_dict = {
    "top": Qt.AlignmentFlag.AlignTop,
    "middle": Qt.AlignmentFlag.AlignVCenter,
    "bottom": Qt.AlignmentFlag.AlignBottom,
}

horizontal_align_dict = {
    "left": Qt.AlignmentFlag.AlignLeft,
    "center": Qt.AlignmentFlag.AlignHCenter,
    "justify": Qt.AlignmentFlag.AlignJustify,
    "right": Qt.AlignmentFlag.AlignRight,
}


class q2label(QLabel, Q2Widget):
    def __init__(self, meta={}):
        super().__init__({"label": meta.get("label", ""), "dblclick": meta.get("dblclick")})
        # super().__init__(meta)
        self.set_text(self.meta["label"])
        # self.setWordWrap(True)
        # self.set_maximum_height(int(self.get_default_height() * 1.5))

    def set_style_sheet(self, style_text):
        super().set_style_sheet(style_text)

        if "vertical-align" in style_text or "text-align" in style_text:
            if isinstance(style_text, str):
                style_dict = {
                    x[0].strip().replace("{", ""): x[1].strip().replace("}", "")
                    for x in [x.split(":") for x in style_text.split(";") if ":" in x]
                }
            else:
                style_dict = style_text
            CA = vertical_align_dict.get(style_dict.get("vertical-align", "top"), Qt.AlignmentFlag.AlignTop)
            CA |= horizontal_align_dict.get(style_dict.get("text-align", "left"), Qt.AlignmentFlag.AlignLeft)

            self.setAlignment(CA)



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sys



from PyQt6.QtWidgets import QDoubleSpinBox

from q2gui.pyqt6.q2widget import Q2Widget
from q2gui.q2utils import int_, float_


class q2doublespin(QDoubleSpinBox, Q2Widget):
    def __init__(self, meta):
        super().__init__(meta)
        self.meta = meta
        self.setDecimals(int_(meta.get("datadec", 1)))
        self.set_text(meta.get("data"))
        self.setSingleStep(0.05)
        if self.meta.get("valid"):
            self.valueChanged.connect(self.meta.get("valid"))
        self.set_fixed_height()

    def set_text(self, text):
        self.setValue(float_(text))

    def get_text(self):
        return self.text().replace(",", ".")



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from PyQt6.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout, QFormLayout, QGridLayout, QApplication
from PyQt6.QtCore import Qt
from q2gui import q2window


q2_align = {
    "": Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignTop,
    "-1": Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignTop,
    "0": Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignTop,
    "1": Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignBottom,
    "2": Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignBottom,
    "3": Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignBottom,
    "4": Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter,
    "5": Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter,
    "6": Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter,
    "7": Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignTop,
    "8": Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignTop,
    "9": Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignTop,
}


def layout(arg="h"):
    if arg.lower().startswith("v"):
        layout = QVBoxLayout()
        layout.setAlignment(q2_align["7"])
    elif arg.lower().startswith("f"):
        layout = QFormLayout()
        layout.setLabelAlignment(q2_align["6"])
        layout.setFormAlignment(q2_align["0"])
        layout.setFieldGrowthPolicy(layout.FieldGrowthPolicy.ExpandingFieldsGrow)
        layout.layout().setSpacing(2)
    elif arg.lower().startswith("g"):
        layout = QGridLayout()
    else:
        layout = QHBoxLayout()
        layout.setAlignment(q2_align["7"])
    layout.layout().setContentsMargins(0, 0, 0, 0)
    # layout.layout().setContentsMargins(3, 3, 3, 3)
    return layout


class Q2Frame(q2window.Q2Frame, QWidget):
    def set_mode(self, mode="v"):
        self.splitter = None
        super().set_mode(mode=mode)
        if self.layout() is not None:
            return
        self.setLayout(layout(mode))

    def insert_widget(self, pos=None, widget=None):
        # if widget:
        #     # widget.setContentsMargins(0,20, 0, 0)
        #     if widget.label:
        #         print("--",widget, widget.label.get_text())
        if hasattr(widget, "meta"):
            stretch_factor = int(widget.meta.get("stretch", 0))
        else:
            stretch_factor = 0
        self.layout().addWidget(widget, stretch=stretch_factor)
        self.updateGeometry()

    def add_row(self, label=None, widget=None):
        # if widget:
        #     widget.setContentsMargins(0, 0, 0, 0)
        #     print("f", widget, widget.label.get_text())
        self.layout().addRow(label, widget)
        self.updateGeometry()


class Q2QtWindow(q2window.Q2Window, Q2Frame):
    def __init__(self, title=""):
        super().__init__()
        self.set_title(title)

    def set_position(self, left, top):
        if left == -9999 and top == -9999:
            self.center_position()
        else:
            self.move(int(left), int(top))

    def center_position(self):
        # sw, sh = (Q_DesktopWidget().size().width(), Q_DesktopWidget().size().height())
        screen = QApplication.screens()[0]

        sw, sh = (screen.size().width(), screen.size().height())

        ww, wh = self.get_size()

        left = int((sw - ww) / 2)
        top = int((sh - wh) / 2)
        self.move(left, top)

    def set_size(self, width, height):
        self.resize(int(width), int(height))

    def get_position(self):
        return (self.pos().x(), self.pos().y())

    def get_size(self):
        if hasattr(self, "parent") and self.parent() is not None:
            return (self.parent().size().width(), self.parent().size().height())
        else:
            return (self.size().width(), self.size().height())

    def set_disabled(self, arg=True):
        self.setEnabled(True if not arg else False)

    def set_enabled(self, arg=True):
        self.setEnabled(True if arg else False)

    def set_title(self, title):
        super().set_title(title)
        QWidget.setWindowTitle(self, title)

    def hide_border(self):
        super().hide_border()
        self.setObjectName("grb")
        # self.setStyleSheet("QGroupBox#grb {border:0}")
        self.parent().setWindowFlags(Qt.WindowType.CustomizeWindowHint | Qt.WindowType.FramelessWindowHint)

    def is_maximized(self):
        return 1 if QWidget.isMaximized(self) else 0

    def show_maximized(self):
        QWidget.showMaximized(self)



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sys

from PyQt6 import QtGui


from q2gui import q2app
from q2gui.q2app import Q2Actions
from q2gui.q2app import GRID_ACTION_TEXT, GRID_ACTION_ICON

from PyQt6.QtWidgets import QFrame, QHBoxLayout, QVBoxLayout, QToolBar, QSizePolicy, QToolButton, QMenu
from PyQt6.QtGui import QIcon, QColor, QFont, QFontMetrics
from PyQt6.QtCore import Qt, QMargins

from q2gui.pyqt6.q2widget import Q2Widget
from q2gui.pyqt6.q2window import q2_align


class q2toolbar(QFrame, Q2Widget):
    def __init__(self, meta):
        super().__init__(meta)
        self.setLayout(QVBoxLayout() if "v" in meta.get("control") else QHBoxLayout())
        # self.setSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Minimum)
        self.layout().setAlignment(q2_align["7"])
        self.layout().setSpacing(-1)
        self.layout().setContentsMargins(QMargins(0, 0, 0, 0))

        self.icon_size = 24
        tmp_icon = q2app.q2_app.get_icon(GRID_ACTION_ICON)
        if tmp_icon:
            self.icon_size = QIcon(tmp_icon).availableSizes()[0].width()

        self.toolBarPanel = QToolBar()
        self.toolBarPanel.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        action_list = []
        if isinstance(meta.get("actions"), Q2Actions):
            action_list.extend(meta.get("actions"))
            actions = meta.get("actions")
        elif isinstance(meta.get("actions"), list):
            for x in meta.get("actions"):
                if isinstance(x, Q2Actions):
                    action_list.extend(x)
            actions = meta.get("actions")[0]

        if action_list is []:
            return

        self.show_main_button = actions.show_main_button
        self.show_actions = actions.show_actions
        tool_bar_qt_actions = QMenu()
        cascade_action = {"": tool_bar_qt_actions}

        for action in action_list:
            if action.get("text", "").startswith("/"):
                continue
            worker = action.get("worker", None)
            action_text_list = action["text"].split("|")
            for x in range(len(action_text_list)):
                action_key = "|".join(action_text_list[:x])
                action_text = action_text_list[x]
                if action_text == "-":
                    action["engineAction"] = cascade_action[action_key].addSeparator()
                else:
                    if x + 1 == len(action_text_list) and (
                        action.get("worker") or (action.get("child_where") and action.get("child_form"))
                    ):  # real action
                        action["engineAction"] = cascade_action[action_key].addAction(action_text)
                        action["parent_action"] = cascade_action[action_key]
                        action["parent_action_text"] = action_key

                        action["_set_visible_parent_action"] = lambda mode=True, act=action[
                            "parent_action"
                        ]: act.setVisible(mode)

                        action["_set_visible"] = lambda mode=True, act=action["engineAction"]: act.setVisible(
                            mode
                        )
                        action["_set_enabled"] = lambda mode=True, act=action["engineAction"]: act.setEnabled(
                            mode
                        )
                        action["_set_disabled"] = lambda mode=True, act=action[
                            "engineAction"
                        ]: act.setDisabled(mode)
                        action["engineAction"].setToolTip(action.get("mess", ""))
                        action["engineAction"].setStatusTip(action.get("mess", ""))
                        action["engineAction"].setObjectName(action.get("tag", ""))

                        action["engineAction"].setIcon(q2app.q2_app.get_engine_icon(action.get("icon", "")))

                        if worker:
                            action["_worker"] = worker
                            action["engineAction"].triggered.connect(worker)
                        elif action.get("child_where") and action.get("child_form"):

                            def getChildForm(action):
                                def rd():
                                    self.meta["form"].show_child_form(action)

                                return rd

                            action["_worker"] = getChildForm(action)
                            action["engineAction"].triggered.connect(getChildForm(action))

                        action["engineAction"].setShortcut(
                            action["hotkey"] if not action["hotkey"] == "Spacebar" else Qt.Key.Key_Space
                        )
                        action["engineAction"].setShortcutContext(
                            Qt.ShortcutContext.WidgetWithChildrenShortcut
                        )
                    else:  # cascade
                        subMenu = "|".join(action_text_list[: x + 1])
                        if subMenu not in cascade_action:
                            cascade_action[subMenu] = cascade_action[action_key].addMenu(
                                f"{action_text}  {'' if '|' in subMenu else '  '}"
                            )
                            if action.get("icon", ""):
                                cascade_action[subMenu].setIcon(
                                    q2app.q2_app.get_engine_icon(action.get("icon", ""))
                                )

        self.main_button = QToolBar()

        self.main_button_action = QToolButton()
        self.main_button_action.setText(GRID_ACTION_TEXT)

        self.main_button_action.setIcon(q2app.q2_app.get_engine_icon(GRID_ACTION_ICON))

        self.main_button_action.setToolTip(self.meta.get("mess", ""))
        self.main_button_action.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)
        self.main_button_action.setMenu(tool_bar_qt_actions)
        self.main_button.addWidget(self.main_button_action)

        self.layout().addWidget(self.main_button)
        if not self.show_main_button:
            self.main_button.setVisible(False)

        self.toolBarPanel.addSeparator()
        self.toolBarPanel.addActions(tool_bar_qt_actions.actions())

        for x in self.toolBarPanel.actions():
            action_widget = self.toolBarPanel.widgetForAction(x)
            if hasattr(action_widget, "setPopupMode"):
                action_widget.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)

        if self.show_actions:
            self.layout().addWidget(self.toolBarPanel)

        self.set_background_color()

    def set_background_color(self):
        color_mode = q2app.q2_app.q2style.get_color_mode()
        if color_mode == "dark":
            background_color = (
                QColor(q2app.q2_app.q2style.get_styles()["background_control"]).lighter(200).name()
            )
        elif color_mode == "light":
            background_color = QColor(q2app.q2_app.q2style.get_styles()["background_control"]).name()
        else:
            background_color = "white"

        for action in self.toolBarPanel.actions():
            color = background_color
            object_name = action.objectName()
            if object_name == "edit":
                object_name = "#41a7fa"
            elif object_name == "select":
                object_name = "#74d484"
            if object_name:
                tmp_color = QColor(object_name).lighter(140).name()
                if tmp_color != "#000000":
                    color = tmp_color
            hover_color = QColor(color).darker(150).name()
            action_widget = self.toolBarPanel.widgetForAction(action)
            action_widget.setStyleSheet(
                """QToolButton {background: %s; margin: 0 2; color:black} 
                    QToolButton:hover {background: %s}
                    QToolButton:disabled {background: %s}
                    """
                % (color, hover_color, background_color)
            )
            if len(action.objectName()) == 1:
                action_widget.setText(action.objectName())
            elif action.property("unicode_icon"):
                action_widget.setText(action.property("unicode_icon"))

    def set_context_menu(self, widget):
        widget.setContextMenuPolicy(Qt.ContextMenuPolicy.ActionsContextMenu)
        widget.addActions(self.toolBarPanel.actions())

    def showEvent(self, ev):
        button_height = self.main_button.sizeHint().height()
        self.toolBarPanel.setMaximumHeight(button_height)
        for x in self.toolBarPanel.actions():
            action_widget = self.toolBarPanel.widgetForAction(x)
            if isinstance(action_widget, QToolButton):
                if action_widget.icon().availableSizes() == []:
                    # no icon
                    font = action_widget.font()
                    # font.setPointSize(font.pointSize()+1)
                    font.setWeight(QFont.Weight.Medium)
                    action_widget.setFont(font)
                    action_widget.setFixedHeight(button_height)

        return super().showEvent(ev)



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sys




from PyQt6.QtWidgets import (
    QTableView,
    QStyledItemDelegate,
    QAbstractItemView,
    QSizePolicy,
)
from PyQt6.QtGui import QPalette, QPainter

from PyQt6.QtCore import Qt, QAbstractTableModel, QVariant

from q2gui.pyqt6.q2window import q2_align
from q2gui.q2utils import int_
from q2gui.q2model import Q2Model
from q2gui.pyqt6.widgets.q2lookup import q2lookup


sort_ascend_char = "▲"
sort_decend_char = "▼"


class q2Delegate(QStyledItemDelegate):
    def paint(self, painter: QPainter, option, index):
        if self.parent().currentIndex().column() == index.column():
            color = option.palette.color(QPalette.ColorRole.AlternateBase).darker(900)
            color.setAlpha(int(color.alpha() / 5))
            painter.fillRect(option.rect, color)
        super().paint(painter, option, index)


class q2grid(QTableView):
    class Q2TableModel(QAbstractTableModel):
        def __init__(self, q2_model):
            super().__init__(parent=None)
            self.q2_model: Q2Model = q2_model
            self._q2_model_refresh = self.q2_model.refresh
            self.q2_model.refresh = self.refresh

        def set_order(self, column):
            self.q2_model.order_column(column)

        def rowCount(self, parent=None):
            return self.q2_model.row_count()

        def columnCount(self, parent=None):
            return self.q2_model.column_count()

        def refresh(self):
            self.beginResetModel()
            self.endResetModel()
            self._q2_model_refresh()

        def data(self, index, role=Qt.ItemDataRole.DisplayRole):
            control = self.q2_model.meta[index.column()].get("control")
            if role == Qt.ItemDataRole.DisplayRole:
                if control == "check":
                    return QVariant()
                else:
                    return QVariant(self.q2_model.data(index.row(), index.column()))
            elif role == Qt.ItemDataRole.TextAlignmentRole:
                return QVariant(q2_align[str(self.q2_model.alignment(index.column()))])
            elif role == Qt.ItemDataRole.CheckStateRole:
                if control == "check":
                    if self.q2_model.data(index.row(), index.column()):
                        return Qt.CheckState.Checked
                    else:
                        return Qt.CheckState.Unchecked
            else:
                return QVariant()

        def headerData(self, col, orientation, role=Qt.ItemDataRole.DisplayRole):
            if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
                sort_char = ""
                if self.q2_model.columns[col] in self.q2_model.order_text:
                    if "desc" in self.q2_model.order_text:
                        sort_char = sort_decend_char
                    else:
                        sort_char = sort_ascend_char
                return sort_char + self.q2_model.headers[col]
            elif orientation == Qt.Orientation.Vertical and role == Qt.ItemDataRole.DisplayRole:
                return QVariant("")
            else:
                return QVariant()

        def flags(self, index):
            control = self.q2_model.meta[index.column()].get("control")
            flags = super().flags(index)
            if control == "check":
                flags |= Qt.ItemFlag.ItemIsUserCheckable
            return flags

    def __init__(self, meta):
        super().__init__()
        self.meta = meta

        self.q2_form = self.meta.get("form")
        self.q2_model = self.q2_form.model

        self.setItemDelegate(q2Delegate(self))
        self.setTabKeyNavigation(False)
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)

        self.horizontalHeader().setSectionsMovable(True)
        self.horizontalHeader().setDefaultAlignment(q2_align["7"])
        self.horizontalHeader().sectionClicked.connect(self.q2_form.grid_header_clicked)
        h = self.fontMetrics().height()
        self.verticalHeader().setMinimumSectionSize(h + h//4)
        self.verticalHeader().setDefaultSectionSize(self.fontMetrics().height())

        self.setAlternatingRowColors(True)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.ActionsContextMenu)
        self.doubleClicked.connect(self.q2_form.grid_double_clicked)
        self.setModel(self.Q2TableModel(self.q2_form.model))

    def currentChanged(self, current, previous):
        super().currentChanged(current, previous)
        self.model().dataChanged.emit(current, previous)
        self.model().dataChanged.emit(previous, current)
        self.q2_form._grid_index_changed(self.currentIndex().row(), self.currentIndex().column())

    def current_index(self):
        return self.currentIndex().row(), self.currentIndex().column()

    def set_focus(self):
        self.setFocus()

    def has_focus(self):
        return self.hasFocus()

    def is_enabled(self):
        return self.isEnabled()

    def row_count(self):
        return self.model().rowCount()

    def column_count(self):
        return self.model().columnCount()

    def set_index(self, row, column=None):
        self.clearSelection()
        if row < 0:
            row = 0
        elif row > self.row_count() - 1:
            row = self.row_count() - 1

        if column is None:
            column = self.currentIndex().column()
        elif column < 0:
            column = 0
        elif column > self.column_count() - 1:
            column = self.column_count() - 1

        self.setCurrentIndex(self.model().index(row, column))

    def keyPressEvent(self, event):
        event.accept()
        # if ev.key() in [Qt.Key.Key_F] and ev.modifiers() == Qt.ControlModifier:
        #     self.searchText()
        # if event.key() in [Qt.Key.Key_Asterisk]:
        if (
            event.text()
            and event.key() not in (Qt.Key.Key_Escape, Qt.Key.Key_Enter, Qt.Key.Key_Return, Qt.Key.Key_Space)
            and self.model().rowCount() >= 1
            and event.modifiers() != Qt.KeyboardModifier.ControlModifier
            and event.modifiers() != Qt.KeyboardModifier.AltModifier
        ):
            lookup_widget = q2_grid_lookup(self, event.text())
            lookup_widget.show(self, self.currentIndex().column())
        else:
            super().keyPressEvent(event)

    def get_selected_rows(self):
        return [x.row() for x in self.selectionModel().selectedRows()]

    def get_columns_headers(self):
        rez = {}
        hohe = self.horizontalHeader()
        for x in range(0, hohe.count()):
            # col_header = hohe.model().headerData(x, Qt.Orientation.Horizontal, Qt.ItemDataRole.DisplayRole)
            col_header = self.q2_model.headers[x]
            rez[col_header] = x
        return rez

    def get_columns_settings(self):
        rez = []
        hohe = self.horizontalHeader()
        for x in range(0, hohe.count()):
            # header = hohe.model().headerData(x, Qt.Orientation.Horizontal, Qt.ItemDataRole.DisplayRole)
            header = self.q2_model.headers[x]
            width = self.columnWidth(x)
            pos = hohe.visualIndex(x)
            rez.append({"name": header, "data": f"{pos}, {width}"})
        return rez

    def set_column_settings(self, col_settings):
        headers = self.get_columns_headers()
        for x in col_settings:
            if "," not in col_settings[x]:
                continue
            column_pos = int_(col_settings[x].split(",")[0])
            column_width = int_(col_settings[x].split(",")[1])
            # column_width = column_width if column_width else 10
            self.setColumnWidth(headers.get(x), column_width)
            old_visual = self.horizontalHeader().visualIndex(int_(headers[x]))
            self.horizontalHeader().moveSection(old_visual, column_pos)
        self.set_index(0, self.horizontalHeader().logicalIndex(0))


class q2_grid_lookup(q2lookup):
    def lookup_list_selected(self):
        self.q2_grid.set_index(self.found_rows[self.lookup_list.currentRow()][0])
        self.close()

    def lookup_search(self):
        self.lookup_list.clear()
        self.found_rows = self.q2_model.lookup(self.q2_model_column, self.lookup_edit.get_text())
        for x in self.found_rows:
            self.lookup_list.addItem(f"{x[1]}")

    def show(self, q2_grid, column):
        self.q2_grid = q2_grid
        self.q2_model_column = column
        self.q2_model = q2_grid.q2_model
        return super().show()

    def set_geometry(self):
        parent = self.parent()
        rect = parent.visualRect(parent.currentIndex())
        rect.moveTop(parent.horizontalHeader().height() + 2)
        rect.moveLeft(parent.verticalHeader().width() + rect.x() + 2)
        pos = rect.topLeft()
        pos = parent.mapToGlobal(pos)
        self.setFixedWidth(parent.width() - rect.x())
        self.move(pos)



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.



from decimal import Decimal
from PyQt6.QtWidgets import QWidget, QSizePolicy
from PyQt6.QtGui import QFontMetrics, QFont
from PyQt6.QtCore import Qt

from q2gui import q2widget
from q2gui.pyqt6.q2window import q2_align


class Q2Widget(QWidget, q2widget.Q2Widget):
    def __init__(self, meta={}):
        super().__init__()
        q2widget.Q2Widget.__init__(self, meta)
        if self.meta.get("margins"):
            self.apply_meta_margins()
        else:
            self.set_content_margins(1)
        if self.meta.get("tag"):
            self.setObjectName(self.meta.get("tag"))

    def apply_meta_margins(self):
        meta_margins = self.meta.get("margins")
        if isinstance(meta_margins, (int, Decimal)):
            meta_margins = [meta_margins]
        if not isinstance(meta_margins, list):
            meta_margins = [0]
        if len(meta_margins) == 1:
            self.set_content_margins(int(meta_margins[0]))
        elif len(meta_margins) == 2:
            self.set_content_margins(int(meta_margins[0]), meta_margins[1])
        elif len(meta_margins) == 3:
            self.set_content_margins(int(meta_margins[0]), meta_margins[1], meta_margins[2])
        else:
            self.set_content_margins(int(meta_margins[0]), meta_margins[1], meta_margins[2], meta_margins[3])

    def mouseDoubleClickEvent(self, event):
        if self.meta.get("dblclick"):
            self.meta.get("dblclick")()
        return super().mouseDoubleClickEvent(event)

    def set_tooltip(self, mess):
        self.setToolTip(mess)

    def set_disabled(self, arg=True):
        self.setEnabled(True if not arg else False)

    def set_enabled(self, arg=True):
        self.setEnabled(True if arg else False)

    def set_text(self, text):
        if hasattr(self, "setText"):
            self.setText(f"{text}")

    def get_text(self):
        if hasattr(self, "text"):
            return self.text()
        return ""

    def set_readonly(self, arg):
        if hasattr(self, "setReadOnly"):
            self.setReadOnly(True if arg else False)

    def is_enabled(self):
        return self.isEnabled()

    def set_visible(self, arg=True):
        self.setVisible(arg)

    def is_visible(self):
        if hasattr(self, "isVisible"):
            return self.isVisible()

    def is_readonly(self):
        if hasattr(self, "isReadOnly"):
            return self.isReadOnly()

    def set_font(self, font_name="", font_size=12):
        self.setFont(QFont(font_name, font_size))

    def set_focus(self):
        self.setFocus()

    def has_focus(self):
        return self.hasFocus()

    def set_maximum_width(self, width, char="O"):
        if self.meta.get("control", "") not in ("radio", "check"):
            if char != "":
                # self.setMaximumWidth(QFontMetrics(self.font()).width(char) * width)
                self.setMaximumWidth(int(QFontMetrics(self.font()).horizontalAdvance(char) * width))
            else:
                self.setMaximumWidth(int(width))

    def set_minimum_width(self, width, char="O"):
        if self.meta.get("control", "") not in ("radio", "check"):
            if char != "":
                self.setMinimumWidth(int(QFontMetrics(self.font()).horizontalAdvance(char) * width))
            else:
                self.setMinimumWidth(int(width))

    def set_fixed_width(self, width, char="O"):
        if self.meta.get("control", "") not in ("radio", "check"):
            if char != "":
                self.setFixedWidth(int(QFontMetrics(self.font()).horizontalAdvance(char) * width))
            else:
                self.setFixedWidth(int(width))

    def set_fixed_height(self, height=1, char="O"):
        if self.meta.get("control", "") not in ("radio", "check"):
            if char != "":
                self.setFixedHeight(int(QFontMetrics(self.font()).height() * height * 1.6))
            else:
                self.setFixedHeight(int(height))

    def set_maximum_len(self, length):
        if hasattr(self, "setMaxLength"):
            return self.setMaxLength(int(length))

    def set_alignment(self, alignment):
        if hasattr(self, "setAlignment"):
            self.setAlignment(q2_align[f"{alignment}"])

    # def valid(self):
    #     if self.meta.get("valid"):
    #         return self.meta.get("valid", lambda: True)()
    #     else:
    #         return True

    # def when(self):
    #     if self.meta.get("when"):
    #         return self.meta.get("when", lambda: True)()
    #     else:
    #         return True

    def set_style_sheet(self, css: str):
        super().set_style_sheet(css)
        self.setStyleSheet(self.style_sheet)

    def add_style_sheet(self, css: str):
        last_style = " ".join([self.styleSheet(), f"; {css}"])
        super().set_style_sheet(last_style)
        self.setStyleSheet(self.style_sheet)

    def get_style_sheet(self):
        return self.styleSheet()

    # def fix_default_height(self):
    #     self.set_maximum_height(self.get_default_height())

    def get_default_height(self):
        return self.sizeHint().height()

    def set_maximum_height(self, height):
        self.setMaximumHeight(int(height))

    # def fix_default_width(self):
    #     self.set_maximum_width(self.get_default_width())

    def get_default_width(self):
        return self.sizeHint().width()

    def set_size_policy(self, horizontal, vertical):
        sp = {
            "fixed": QSizePolicy.Policy.Fixed,
            "minimum": QSizePolicy.Policy.Minimum,
            "maximum": QSizePolicy.Policy.Maximum,
            "preffered": QSizePolicy.Policy.Preferred,
            "expanding": QSizePolicy.Policy.Expanding,
            "minimalexpanding": QSizePolicy.Policy.MinimumExpanding,
            "ignored": QSizePolicy.Policy.Ignored,
        }

        self.setSizePolicy(
            sp.get(horizontal, QSizePolicy.Policy.Minimum), sp.get(vertical, QSizePolicy.Policy.Minimum)
        )

    def can_get_focus(self):
        if self.focusPolicy() == Qt.FocusPolicy.NoFocus:
            return False
        else:
            return True

    def set_content_margins(self, top=0, right=None, bottom=None, left=None):
        if right is None:
            right = top
        if bottom is None:
            bottom = top
        if left is None:
            left = right
        self.setContentsMargins(top, right, bottom, left)

    def get_next_focus_widget(self, pos=1):
        return self.nextInFocusChain()

    def get_next_widget(self, pos=1):
        return self.layout().widget()

    def add_widget_above(self, widget, pos=0):
        my_pos = self.parentWidget().layout().indexOf(self)
        self.parent().layout().insertWidget(my_pos - pos, widget)

    def add_widget_below(self, widget, pos=0):
        if pos == -1:
            self.parent().layout().addWidget(widget)
        else:
            my_pos = self.parentWidget().layout().indexOf(self)
            self.parent().layout().insertWidget(my_pos + pos + 1, widget)

    def remove(self):
        self.parentWidget().layout().removeWidget(self)
        self.setParent(None)

    def get_layout_position(self):
        return self.parentWidget().layout().indexOf(self)

    def get_layout_count(self):
        return self.parentWidget().layout().count()

    def get_layout_widget(self, pos):
        return self.parentWidget().layout().itemAt(pos).widget()

    def get_layout_widgets(self):
        return [self.get_layout_widget(x) for x in range(self.get_layout_count())]

    def move_up(self):
        pos = self.get_layout_position()
        if pos > 0:
            w = self.parentWidget().layout().takeAt(pos).widget()
            self.parentWidget().layout().insertWidget(pos - 1, w)

    def move_down(self):
        pos = self.get_layout_position()
        if pos < self.get_layout_count() - 1:
            w = self.parentWidget().layout().takeAt(pos + 1).widget()
            self.parentWidget().layout().insertWidget(pos, w)

    def action_set_visible(self, text, mode=True):
        for action in self.actions():
            if action.text().strip() == text:
                action.setVisible(mode)

    def action_set_enabled(self, text, mode=True):
        for action in self.actions():
            if action.text().strip() == text:
                action.setVisible(mode)



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sys



# from PyQt6.QtWidgets import QTextEdit, QSizePolicy
from PyQt6.QtWidgets import (
    QVBoxLayout,
    QHBoxLayout,
    QWidget,
)

from PyQt6.Qsci import QsciScintilla, QsciLexerPython, QsciLexerSQL, QsciLexerJSON, QsciAPIs, QsciLexer
from PyQt6.QtGui import QColor

# from PyQt6.QtWidgets import QMenu
from PyQt6.QtCore import Qt, QTimer, QSize

from q2gui.pyqt6.q2widget import Q2Widget
from q2gui.pyqt6.widgets.q2line import q2line
from q2gui.pyqt6.widgets.q2label import q2label
from q2gui.pyqt6.widgets.q2button import q2button
from q2gui.q2utils import int_


class q2code(QsciScintilla, Q2Widget):
    def __init__(self, meta):
        super().__init__(meta)
        self.setUtf8(True)
        self.setFolding(QsciScintilla.FoldStyle.BoxedTreeFoldStyle)

        self.lexer: QsciLexer = None
        self.set_lexer()
        self.set_background_color()

        self.setAutoIndent(True)
        self.setIndentationGuides(True)
        self.setIndentationsUseTabs(False)
        self.setBraceMatching(QsciScintilla.BraceMatch.StrictBraceMatch)
        self.setMarginLineNumbers(1, True)
        self.setMarginWidth(1, "9999")
        self.setTabWidth(4)

        self.setAutoCompletionSource(QsciScintilla.AutoCompletionSource.AcsAll)
        self.setAutoCompletionCaseSensitivity(True)
        self.setAutoCompletionReplaceWord(True)
        self.setAutoCompletionThreshold(1)

        self.SCN_DOUBLECLICK

        self.setCaretLineVisible(True)

        self.searchIndicator = QsciScintilla.INDIC_CONTAINER
        self.SendScintilla(QsciScintilla.SCI_INDICSETSTYLE, self.searchIndicator, QsciScintilla.INDIC_BOX)
        self.SendScintilla(QsciScintilla.SCI_INDICSETFORE, self.searchIndicator, QColor("red"))

        self.cursorPositionChanged.connect(self.__cursorPositionChanged)
        self.__markOccurrencesTimer = QTimer(self)
        self.__markOccurrencesTimer.setSingleShot(True)
        self.__markOccurrencesTimer.setInterval(500)
        self.__markOccurrencesTimer.timeout.connect(self.__markOccurrences)
        self.textChanged.connect(self.valid)
        # self.setContextMenuPolicy(Qt.ContextMenuPolicy.ActionsContextMenu)
        self.editor_panel = q2editor_panel(self)
        # self.create_context_menu()
        self.set_custom_autocompletition_list()

    def set_lexer(self, lexer=""):
        if lexer == "":
            lexer = self.meta["control"]
        if "python" in lexer:
            self.lexer = QsciLexerPython()
        elif "sql" in lexer:
            self.lexer = QsciLexerSQL()
        elif "json" in lexer:
            self.lexer = QsciLexerJSON()
        else:
            self.lexer = QsciLexerPython()
        if self.lexer:
            self.setLexer(self.lexer)

    def set_custom_autocompletition_list(self, custom_autocompletions_list=[]):
        custom_autocompletions_list = self.meta["form"].q2_app.get_autocompletition_list()
        self.__api = QsciAPIs(self.lexer)
        for ac in custom_autocompletions_list:
            self.__api.add(ac)
        self.__api.prepare()

    def set_background_color(self, red=150, green=200, blue=230):
        # color_mode = self.meta.get("q2_app").q2style.get_system_color_mode()
        # print(color_mode, self.meta.get("q2_app").q2style.color_mode)
        # b_color = QColor(self.meta.get("q2_app").q2style.styles.get(color_mode).get("background"))
        # f_color = QColor(self.meta.get("q2_app").q2style.styles.get(color_mode).get("color"))

        # self.setMatchedBraceForegroundColor(QColor("lightgreen"))
        self.lexer.setDefaultPaper(QColor(red, green, blue))
        self.lexer.setPaper(QColor(red, green, blue))
        # self.lexer.setDefaultPaper(b_color)
        # self.lexer.setPaper(b_color)
        # self.lexer.setColor(f_color, 5)
        self.setMarginsForegroundColor(QColor("black"))
        self.setMarginsBackgroundColor(QColor("gray"))

    def __cursorPositionChanged(self, line, index):
        self.__markOccurrencesTimer.stop()
        self.clearIndicatorRange(
            0, 0, self.lines() - 1, len(self.text(self.lines() - 1)), self.searchIndicator
        )
        self.__markOccurrencesTimer.start()

    def __findFirstTarget(self, text):
        if text == "":
            return False
        self.__targetSearchExpr = text.encode("utf-8")
        self.__targetSearchFlags = QsciScintilla.SCFIND_MATCHCASE | QsciScintilla.SCFIND_WHOLEWORD
        self.__targetSearchStart = 0
        self.__targetSearchEnd = self.SendScintilla(QsciScintilla.SCI_GETTEXTLENGTH)
        self.__targetSearchActive = True
        return self.__doSearchTarget()

    def __findNextTarget(self):
        if not self.__targetSearchActive:
            return False
        return self.__doSearchTarget()

    def __doSearchTarget(self):
        if self.__targetSearchStart == self.__targetSearchEnd:
            self.__targetSearchActive = False
            return False
        self.SendScintilla(QsciScintilla.SCI_SETTARGETSTART, self.__targetSearchStart)
        self.SendScintilla(QsciScintilla.SCI_SETTARGETEND, self.__targetSearchEnd)
        self.SendScintilla(QsciScintilla.SCI_SETSEARCHFLAGS, self.__targetSearchFlags)
        pos = self.SendScintilla(
            QsciScintilla.SCI_SEARCHINTARGET, len(self.__targetSearchExpr), self.__targetSearchExpr
        )
        if pos == -1:
            self.__targetSearchActive = False
            return False
        self.SendScintilla(QsciScintilla.SCI_INDICATORFILLRANGE, pos, len(self.__targetSearchExpr))
        targend = self.SendScintilla(QsciScintilla.SCI_GETTARGETEND)
        self.__targetSearchStart = targend
        return True

    def __markOccurrences(self):
        if self.hasFocus():
            line, index = self.getCursorPosition()
            ok = self.__findFirstTarget(self.__getWord(self.text(line), index - 1))
            while ok:
                ok = self.__findNextTarget()

    def __getWord(self, text, index):
        word = ""
        for x in range(index, -1, -1):
            if text[x].isalpha() or text[x].isdigit():
                word = text[x] + word
            else:
                break
        for x in range(index + 1, len(text)):
            if text[x].isalpha() or text[x].isdigit():
                word += text[x]
            else:
                break
        return word

    def mouseDoubleClickEvent(self, event):
        return super().mouseDoubleClickEvent(event)

    def focusInEvent(self, ev):
        super().focusInEvent(ev)
        self.create_context_menu()

    def focusOutEvent(self, ev):
        self.clear_actions()
        super().focusInEvent(ev)

    def clear_actions(self):
        for x in self.actions():
            self.removeAction(x)

    def create_context_menu(self):
        self.context_menu = self.createStandardContextMenu()
        self.context_menu.addSeparator()

        find_action = self.context_menu.addAction("Find next")
        find_action.triggered.connect(self.editor_panel.show_find_next)
        find_action.setShortcuts(["Ctrl+F", "F3"])

        find_action = self.context_menu.addAction("Find prev")
        find_action.triggered.connect(self.editor_panel.show_find_prev)
        find_action.setShortcuts(["Shift+F3"])

        replace_action = self.context_menu.addAction("Replace")
        replace_action.triggered.connect(self.editor_panel.show_replace)
        replace_action.setShortcut("Ctrl+H")

        self.context_menu.addSeparator()

        move_up_action = self.context_menu.addAction("Move selection up")
        move_up_action.triggered.connect(self.perform_move_up)
        move_up_action.setShortcuts(["Ctrl+Alt+Up"])

        move_down_action = self.context_menu.addAction("Move selectipn down")
        move_down_action.triggered.connect(self.perform_move_down)
        move_down_action.setShortcuts(["Ctrl+Alt+Down"])

        self.context_menu.addSeparator()

        gotoline_action = self.context_menu.addAction("Go to line")
        gotoline_action.triggered.connect(self.editor_panel.show_go)
        gotoline_action.setShortcut("Ctrl+G")

        self.context_menu.addSeparator()

        fold_action = self.context_menu.addAction("Fold/Unfold")
        fold_action.triggered.connect(self.perform_folding)
        fold_action.setShortcuts(["Alt+Up", "Alt+Down"])

        foldall_action = self.context_menu.addAction("Fold/Unfold All")
        foldall_action.triggered.connect(self.foldAll)
        foldall_action.setShortcuts(["Alt+Left", "Alt+Right"])

        self.context_menu.addSeparator()
        comment_action = self.context_menu.addAction("Comment/uncomment line(s)")
        comment_action.setShortcut("Ctrl+3")
        comment_action.triggered.connect(self.perform_comment)

        complete_action = self.context_menu.addAction("Autocomplete")
        complete_action.triggered.connect(self.autoCompleteFromAll)
        complete_action.setShortcuts(["Ctrl+Space"])

        self.addActions(self.context_menu.actions())
        for x in self.context_menu.actions():
            if x.isEnabled():
                x.setVisible(True)
            else:
                x.setVisible(False)

    def perform_move_down(self):
        self.SendScintilla(QsciScintilla.SCI_MOVESELECTEDLINESDOWN)

    def perform_move_up(self):
        self.SendScintilla(QsciScintilla.SCI_MOVESELECTEDLINESUP)

    def perform_comment(self):
        selected_lines = []
        current_line, current_pos = self.getCursorPosition()
        if self.hasSelectedText():
            line1, pos1, line2, pos2 = self.getSelection()
            for x in range(line1, line2 + 1):
                selected_lines.append(x)
        else:
            selected_lines.append(current_line)

        for line in selected_lines:
            new_pos = self.comment_line(line, current_pos)
            if line == current_line:
                self.setCursorPosition(current_line, new_pos)

        if len(selected_lines) > 1:
            self.setSelection(line1, pos1, line2, pos2)

    def comment_line(self, line, pos):
        current_line = self.text(line)
        if len(current_line.strip()):
            if current_line.lstrip()[0:2] == "# ":
                comment_pos = current_line.index("# ")
                if comment_pos < pos:
                    pos -= 2
                self.setSelection(line, comment_pos, line, comment_pos + 2)
                self.removeSelectedText()
            else:
                spaces_count = len(current_line) - len(current_line.lstrip())
                if pos > spaces_count:
                    pos += 2
                self.insertAt("# ", line, spaces_count)
        return pos

    def perform_folding(self):
        self.foldLine(self.getCursorPosition()[0])

    def current_line(self):
        return self.getCursorPosition()[0]

    def goto_line(self, line=0):
        self.setCursorPosition(int_(line) - 1, 0)
        self.set_focus()
        self.ensureLineVisible(self.getCursorPosition()[0])
        self.editor_panel.close()

    def contextMenuEvent(self, event):
        self.create_context_menu()
        self.context_menu.exec(event.globalPos())

    def showEvent(self, ev):
        self.updateGeometry()
        return super().showEvent(ev)

    def sizeHint(self):
        if self.isVisible():
            return QSize(99999, 99999)
        else:
            return super().sizeHint()


class q2editor_panel(QWidget):
    def __init__(self, parent, text=""):
        super().__init__(parent, Qt.WindowType.Popup)
        # super().__init__(parent)
        self.setLayout(QVBoxLayout())
        self.setObjectName("editor_panel")

        self.layout().setContentsMargins(0, 0, 0, 0)
        # self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        self.create_find()
        self.create_replace()
        self.create_go()

        self.layout().addWidget(self.find_frame)
        self.layout().addWidget(self.replace_frame)
        self.layout().addWidget(self.go_frame)

        self.set_find_text(text)
        self.q2code: q2code = parent
        self.last_find_direction = "down"

    def create_find(self):
        self.find_frame = QWidget()

        self.find_frame.setLayout(QHBoxLayout())
        self.find_frame.layout().setContentsMargins(0, 0, 0, 0)

        self.find_next_button = q2button({"label": ">"})
        self.find_next_button.set_fixed_width(3, "w")

        self.find_prev_button = q2button({"label": "<"})
        self.find_prev_button.set_fixed_width(3, "w")

        self.find_text = q2line({})
        self.find_frame.layout().addWidget(q2label({"label": "find:"}))
        self.find_frame.layout().addWidget(self.find_text)
        self.find_frame.layout().addWidget(self.find_prev_button)
        self.find_frame.layout().addWidget(self.find_next_button)
        self.find_next_button.pressed.connect(self.perform_find_next)
        self.find_prev_button.pressed.connect(self.perform_find_prev)
        self.find_text.keyPressEvent = self.find_keyPressEvent

    def find_keyPressEvent(self, event):
        if event.key() in [Qt.Key.Key_Up] or (
            event.key() in [Qt.Key.Key_Enter, Qt.Key.Key_Return, Qt.Key.Key_F3]
            and event.modifiers() == Qt.KeyboardModifier.ShiftModifier
        ):
            self.perform_find_prev()
        elif event.key() in [Qt.Key.Key_Enter, Qt.Key.Key_Return, Qt.Key.Key_Down, Qt.Key.Key_F3]:
            self.perform_find_next()
        elif event.key() == Qt.Key.Key_H:
            self.show_replace()
        q2line.keyPressEvent(self.find_text, event)

    def perform_find_next(self):
        self.q2code.findFirst(self.find_text.get_text(), False, False, False, False, True)
        self.last_find_direction = "down"

    def perform_find_prev(self):
        line1, index1, line2, index2 = self.q2code.getSelection()
        self.q2code.setCursorPosition(line1, index1)
        self.q2code.findFirst(self.find_text.get_text(), False, False, False, False, False)
        self.last_find_direction = "up"

    def create_replace(self):
        self.replace_frame = QWidget()
        self.replace_frame.setLayout(QHBoxLayout())
        self.replace_frame.layout().setContentsMargins(0, 0, 0, 0)
        self.replace_text = q2line({})
        self.replace_text.returnPressed.connect(self.perform_replace)

        self.replace_frame.layout().addWidget(q2label({"label": "replace:"}))
        self.replace_frame.layout().addWidget(self.replace_text)

        self.replace_button = q2button({"label": "Ok", "datalen": 4})
        self.replace_button.pressed.connect(self.perform_replace)
        # self.replace_button.set_fixed_width(3, "w")
        self.replace_frame.layout().addWidget(self.replace_button)

    def perform_replace(self):
        self.q2code.replace(self.replace_text.get_text())
        if self.last_find_direction == "down":
            self.perform_find_next()
        else:
            self.perform_find_prev()

    def create_go(self):
        self.go_frame = QWidget()
        self.go_frame.setLayout(QHBoxLayout())
        self.go_frame.layout().setContentsMargins(0, 0, 0, 0)
        self.go_text = q2line({"datatype": "int", "datalen": 5, "num": 1})
        self.go_frame.layout().addWidget(q2label({"label": "go to line:"}))
        self.go_frame.layout().addWidget(self.go_text)
        self.go_button = q2button({"label": ">"})
        self.go_button.set_fixed_width(3, "w")
        self.go_frame.layout().addWidget(self.go_button)
        self.go_frame.layout().addStretch()
        self.go_text.editingFinished.connect(self.perform_go)
        self.go_button.pressed.connect(self.perform_go)

    def perform_go(self):
        self.q2code.goto_line(self.go_text.get_text())

    def show_find(self):
        st = self.q2code.selectedText()
        if st:
            self.find_text.set_text(st)
        self.find_frame.show()
        self.replace_frame.hide()
        self.go_frame.hide()
        self.find_text.set_focus()
        self.show()

    def show_find_prev(self):
        self.show_find()
        self.perform_find_prev()

    def show_find_next(self):
        self.show_find()
        self.perform_find_next()

    def show_replace(self):
        self.find_frame.show()
        self.replace_frame.show()
        self.go_frame.hide()
        self.show()

    def show_go(self):
        self.find_frame.hide()
        self.replace_frame.hide()
        self.go_frame.show()
        self.go_text.set_alignment(9)
        self.go_text.set_text("")
        self.go_text.set_focus()
        self.show()

    def set_find_text(self, text=""):
        self.find_text.set_text(text)
        self.find_text.set_focus()

    def show(self):
        self.set_geometry()
        return super().show()

    def set_geometry(self):
        parent = self.q2code
        rect = parent.rect()
        self.setFixedWidth(400 if rect.width() > 400 else int(rect.width() / 2))
        pos = rect.topLeft()
        pos.setX(rect.width() - self.width())
        pos = parent.mapToGlobal(pos)
        self.move(pos)



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sys



from PyQt6.QtWidgets import QTabBar, QTabWidget
from PyQt6.QtGui import QKeySequence, QShortcut
from PyQt6.QtCore import Qt

from q2gui.pyqt6.q2window import Q2Frame
from q2gui.pyqt6.q2widget import Q2Widget
from q2gui.q2utils import int_


class Q2TabBar(QTabBar, Q2Widget):
    def get_text(self):
        return self.tabText(self.currentIndex())


class q2tab(QTabWidget, Q2Widget, Q2Frame):
    def __init__(self, meta):
        super().__init__(meta)
        self.setTabBar(Q2TabBar())
        self.meta = meta

        self.set_tabbar_position(meta.get("alignment", 7))

        self.next_tab_hotkey = QShortcut(QKeySequence("Ctrl+PgDown"), self)
        self.next_tab_hotkey.activated.connect(self.next_tab)

        self.prev_tab_hotkey = QShortcut(QKeySequence("Ctrl+PgUp"), self)
        self.prev_tab_hotkey.activated.connect(self.prev_tab)

    def set_tabbar_position(self, pos):
        pos = str(pos)
        qss = ""
        if "4" in pos:  # left
            self.setTabPosition(QTabWidget.TabPosition.West)
            if "7" in pos:
                qss = "left"
            elif "1" in pos:
                qss = "right"
            else:
                qss = "center"
        elif "2" in pos:  # bottom
            self.setTabPosition(QTabWidget.TabPosition.South)
            if "1" in pos:
                qss = "left"
            elif "3" in pos:
                qss = "right"
            else:
                qss = "center"
        elif "6" in pos:  # right
            self.setTabPosition(QTabWidget.TabPosition.East)
            if "9" in pos:
                qss = "left"
            elif "3" in pos:
                qss = "right"
            else:
                qss = "center"
        else:
            if "9" in pos:
                qss = "right"
            elif "8" in pos:
                qss = "center"
        if qss:
            self.setStyleSheet("QTabWidget::tab-bar { alignment: %s;}" % qss)

    def next_tab(self):
        self.setCurrentIndex(self.currentIndex() + 1)

    def prev_tab(self):
        self.setCurrentIndex(self.currentIndex() - 1)

    def set_enabled(self, mode=True):
        self.tabBar().setEnabled(mode)

    def set_disabled(self, mode=True):
        self.tabBar().setDisabled(mode)

    def set_tab(self, index=0):
        if isinstance(index, str):
            for x in range(self.tabBar().count()):
                print(self.tabBar().tabText(x))
                if self.tabBar().tabText(x) == index:
                    index = x
                    break
        else:
            index = int_(index)
            if index >= self.count():
                index = self.count() - 1
            elif index < 0:
                index = 0
        self.setCurrentIndex(index)

    def minimumSizeHint(self):
        self.setMinimumHeight(super().minimumSizeHint().height())
        return super().minimumSizeHint()

    def set_shortcuts_local(self):
        self.next_tab_hotkey.setContext(Qt.ShortcutContext.WidgetWithChildrenShortcut)
        self.prev_tab_hotkey.setContext(Qt.ShortcutContext.WidgetWithChildrenShortcut)

    def add_tab(self, widget, text=""):
        self.addTab(widget, text)

    def get_text(self):
        return self.tabBar().tabText(self.currentIndex())



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sys



from PyQt6.QtWidgets import QFrame
from PyQt6.QtCore import Qt

import q2gui.q2app as q2app
from q2gui.q2form import Q2Form
from q2gui.q2utils import num
from q2gui.pyqt6.q2widget import Q2Widget
from q2gui.pyqt6.q2window import Q2Frame
from q2gui.pyqt6.widgets.q2line import q2line
from q2gui.pyqt6.widgets.q2button import q2button
from q2gui.pyqt6.widgets.q2lookup import q2lookup
from q2gui.q2model import Q2Model
from q2db.db import Q2Db


class q2relation(QFrame, Q2Widget, Q2Frame):
    def __init__(self, meta):
        super().__init__(meta)
        Q2Frame.__init__(self, "h")
        self.meta = meta

        self.get = q2line(meta)
        self.filter_string = ""

        self.get.textChanged.connect(self.get_text_changed)
        self.get.editingFinished.connect(self.edit_done)
        self.button = q2button(
            {
                "label": "?",
                "datalen": 3,
                "valid": self.show_related_form,
                # "when": self.when,
                "form": self.meta["form"],
            }
        )
        self.say = q2line({"disabled": "*"})
        self.to_form = None
        if self.meta.get("to_form"):
            if isinstance(self.meta.get("to_form"), Q2Form):
                self.to_form: Q2Form = self.meta.get("to_form")
            else:
                self.to_form: Q2Form = self.meta.get("to_form")()
            # self.to_form.max_child_level = 0
            self.to_form.title += " ."

        self.add_widget(self.get)
        self.add_widget(self.button)
        self.add_widget(self.say)
        self.set_text(self.meta.get("data", ""))
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        self.setFocusProxy(self.get)
        # self.get_valid()

    def show_related_form(self):
        if isinstance(self.to_form, Q2Form):
            self.to_form.add_action(
                q2app.ACTION_SELECT_TEXT,
                self.show_related_form_result,
                hotkey="Enter",
                tag="select",
                icon=q2app.ACTION_SELECT_ICON,
            )

            def seek():
                if self.filter_string:
                    self.to_form.model.cursor.set_where(self.filter_string)
                    self.to_form.model.cursor.refresh()

                row = self.to_form.model.cursor.seek_row({self.meta["to_column"]: self.get_text()})
                self.to_form.set_grid_index(row)

            self.to_form.before_grid_show = seek
            self.to_form.show_mdi_modal_grid()
            # self.set_related()

    def show_related_form_result(self):
        if self.to_form:
            self.get.set_text(self.to_form.r.__getattr__(self.meta["to_column"]))
            self.to_form.close()
            self.get.set_focus()
            self.get_valid()

    def edit_done(self):
        self.get_valid()

    def get_valid(self):
        value = self.get.get_text()
        if self.meta.get("num") and num(value) == 0:
            return True
        elif value == "":
            return True
        return self.set_related()

    def get_related(self):
        rel = None
        if self.meta["form"].model:
            rel = self.meta["form"].model._get_related(
                self.get.text(), self.meta, do_not_show_value=1, reset_cache=1
            )
        elif self.meta["form"].db:  # datasource provided
            rel = self.meta["form"].db.get(
                self.meta["to_table"],
                f"{self.meta['to_column']}='{self.get.text()}'",
                self.meta["related"],
            )
        return rel

    def set_related(self):
        rel = self.get_related()
        if rel == "":
            rel = None
            self.say.set_text("")
            self.show_related_form()
            return True

        if rel is None:
            self.say.set_text("")
            return False
        else:
            self.say.set_text(rel)
            return True

    def set_text(self, text):
        if hasattr(self, "get"):
            self.get.set_text(text)
            self.get_valid()

    def get_text(self):
        return self.get.get_text()

    def get_text_changed(self, text):
        if "*" in text and self.meta.get("num") or text.startswith("*"):
            lookup_widget = q2_realtion_lookup(self, "")
            lookup_widget.show(self.meta)

    def set_focus(self):
        self.get.setFocus()

    def set_disabled(self, arg=True):
        if hasattr(self, "get"):
            self.get.set_disabled(arg)
            self.button.set_disabled(arg)
            # self.set_disabled(arg)

    def set_enabled(self, arg=True):
        if hasattr(self, "get"):
            self.get.set_enabled(arg)
            self.button.set_enabled(arg)
            # self.set_enabled(arg)

    def is_enabled(self):
        return self.get.is_enabled()


class q2_realtion_lookup(q2lookup):
    def __init__(self, parent, text):
        super().__init__(parent, text)

    def lookup_search(self):
        self.lookup_list.clear()
        if self.parent().to_form:
            q2_db: Q2Db = self.parent().to_form.model.cursor.q2_db
        else:
            q2_db: Q2Db = self.parent().meta["form"].db

        sql = "select {to_column} as tocol, {related} as recol from {to_table}".format(**self.meta)

        related = self.meta.get("related")

        # cond_list = self.parent().to_form.model.parse_lookup_text(self.lookup_edit.get_text())
        cond_list = Q2Model().parse_lookup_text(self.lookup_edit.get_text())
        where = " and ".join(
            [f"{related} {'' if x[0] == '+' else 'not'} like '%{x[1]}%' " for x in cond_list]
        )
        if where:
            sql += f" where {where}"
        cursor = q2_db.cursor(sql)
        self.found_rows = [x for x in cursor.records()]
        for x in self.found_rows:
            self.lookup_list.addItem(f"""{x["recol"]} ({x["tocol"]})""")

    def lookup_list_selected(self):
        self.parent().set_text(self.found_rows[self.lookup_list.currentRow()]["tocol"])
        self.close()

    def set_geometry(self):
        parent = self.parent()
        self.setFixedWidth(parent.width())
        self.move(parent.mapToGlobal(parent.rect().topLeft()))

    def show(self, meta):
        self.meta = meta
        return super().show()
